//Data-path
//Kamyar Rafati & Naeem Esfahani

module Pipelined_MIPS(reset, clk);
	input		reset, clk;
	wire[31:0]	PCcurrent, PCwriteBack, NextInstPC, IF_Inst, PCcalculated, SignExtended, ShiftedLeft,
				SelectedRegDataA, SelectedRegDataB, EX_SignExtended, EX_DataB,
				EX_DataA, EX_Inst, MEM_ALUResult, EX_ALUResult, EX_MUX_DataA, EX_MUX_DataB, 
				MEM_Data, MEM_ReadData, WB_ReadData, EX_SrcMUX_DataB, ID_JumpAddr, ID_BranchAddr,
				WB_Address, ID_Inst, ID_PC, RegFileA, RegFileB, WB_WriteData;
	wire[7:0]	ID_Control, ID_CorrectedControl;
	wire[4:0]	WB_WriteAddr, EX_Rd, MEM_Rd;
	wire[2:0]	EX_Opcode;
	wire[1:0]	ForwardEqualA, ForwardEqualB, EX_ALUOp, EX_ForwardMuxA, EX_ForwardMuxB;
	wire		PCenable, IFID_WriteEnable, IF_Flush, RegDataBranch, RegDataEqual, BranchInvert,
				IDEX_FlushSelect, WB_RegWrite, Branch, EX_MemToReg, EX_RegWrite, EX_MemRead, EX_ALUSrc, EX_RegDst,
				PCSelect, EX_MemWrite, EX_AluSrc, PCJump;
	
	// Instruction Fetch
	Memory			IRMemory (PCcurrent, 32'b0, IF_Inst, 1'b0, 1'b1, clk);
	Reg_32bit		PC (PCwriteBack, PCcurrent, PCenable, reset, clk);
	Add 			PCadder (PCcurrent, 32'b100, NextInstPC);
	MUX2x1_32bit	IF_PCMux (PCSelect, NextInstPC, ID_BranchAddr, PCcalculated);
	and 			BranchAnd (PCSelect, RegDataBranch, Branch);
	or 				FlushOr (IF_Flush, reset, PCSelect, PCJump);
	MUX2x1_32bit	IF_JumpMux (PCJump, PCcalculated, ID_JumpAddr, PCwriteBack);
	
	// IF/ID reg
	Reg_32bit	IFIDPC(NextInstPC, ID_PC, IFID_WriteEnable, IF_Flush, clk);
	Reg_32bit	IFIDInst(IF_Inst, ID_Inst, IFID_WriteEnable, IF_Flush, clk);

	//Instruction Decode
	Hazard_Detect		HazardDetectUnit(PCenable, ID_Inst[20:16], ID_Inst[25:21], EX_Inst[20:16], EX_MemRead, 
									IFID_WriteEnable, PCenable, IDEX_FlushSelect, reset, clk);
	Control 			ControlUnit(ID_Inst[31:26], Branch, BranchInvert, ID_Control, PCJump, reset, clk);
	Add 				PCcalc(ShiftedLeft, ID_PC, ID_BranchAddr);

	xor 				BranchInverter (RegDataBranch, BranchInvert, RegDataEqual);
	MUX2x1_8bit			IDEX_Flush(IDEX_FlushSelect, ID_Control, 8'b0, ID_CorrectedControl);
	ShiftLeft2_32bit 	SiftLeft(SignExtended, ShiftedLeft);
	RegFile_32bit 		RegFile(ID_Inst[25:21], RegFileA, ID_Inst[20:16], RegFileB, WB_WriteAddr, 
						WB_WriteData, WB_RegWrite, clk);
	SignExtend_32bit 	SignExtend (ID_Inst[15:0], SignExtended);
	MUX3x1_32bit 		CompareMuxA(ForwardEqualA, RegFileA, MEM_ALUResult, EX_ALUResult, SelectedRegDataA);
	MUX3x1_32bit 		CompareMuxB(ForwardEqualB, RegFileB, MEM_ALUResult, EX_ALUResult, SelectedRegDataB);
	RegFile_Compare 	RegCompare(SelectedRegDataA, SelectedRegDataB, RegDataEqual, nil, clk);
	JumpBox 			MyJumpBox(ID_PC[31:28], ID_Inst[25:0], ID_JumpAddr);
	
	// ID/EX reg
	Reg_32bit	IDEXInst(ID_Inst, EX_Inst, 1'b1, reset, clk);
	Reg_32bit	DataA(RegFileA, EX_DataA, 1'b1, reset, clk);
	Reg_32bit	DataB(RegFileB, EX_DataB, 1'b1, reset, clk);
	Reg_32bit	IDEX_SignExtended(SignExtended, EX_SignExtended, 1'b1, reset, clk);
	Reg_1bit	IDEX_MemToReg(ID_CorrectedControl[6], EX_MemToReg, 1'b1, reset, clk);
	Reg_1bit	IDEX_RegWrite(ID_CorrectedControl[7], EX_RegWrite, 1'b1, reset, clk);
	Reg_1bit	IDEX_MemWrite(ID_CorrectedControl[4], EX_MemWrite, 1'b1, reset, clk);
	Reg_1bit	IDEX_MemRead(ID_CorrectedControl[5], EX_MemRead, 1'b1, reset, clk);
	Reg_1bit	IDEX_ALUSrc(ID_CorrectedControl[3], EX_AluSrc, 1'b1, reset, clk);
	Reg_1bit	IDEX_RegDst(ID_CorrectedControl[0], EX_RegDst, 1'b1, reset, clk);
	Reg_2bit	IDEX_ALUOp(ID_CorrectedControl[2:1], EX_ALUOp, 1'b1, reset, clk);

	// Execute
	Forward_Unit	Forward (EX_Inst[20:16], EX_Inst[25:21], EX_Rd,  ID_Inst[20:16], ID_Inst[25:21], 
							MEM_Rd, WB_WriteAddr, 
							MEM_RegWrite, WB_RegWrite, EX_RegWrite, EX_ForwardMuxA, 
							EX_ForwardMuxB, ForwardEqualA, ForwardEqualB, EX_MemRead);
	MUX3x1_32bit	ALUMUXForwardA(EX_ForwardMuxA, EX_DataA, WB_WriteData, MEM_ALUResult, EX_MUX_DataA);
	MUX3x1_32bit	ALUMUXForwardB(EX_ForwardMuxB, EX_DataB, WB_WriteData, MEM_ALUResult, EX_MUX_DataB);
	MUX2x1_32bit	ALUMUXSourceB(EX_AluSrc, EX_MUX_DataB, EX_SignExtended, EX_SrcMUX_DataB);
	ALU_mips		ALU(EX_MUX_DataA, EX_SrcMUX_DataB, EX_Opcode, EX_ALUResult, nil);
	MUX2x1_5bit		EX_RdMUX(EX_RegDst, EX_Inst[20:16], EX_Inst[15:11], EX_Rd);
	ALU_control		EX_ALUControl(EX_ALUOp, EX_SignExtended[5:0], EX_Opcode);
	
	// EX/MEM reg
	Reg_32bit	EXMEM_ALUResult(EX_ALUResult, MEM_ALUResult, 1'b1, reset, clk);
	Reg_32bit	EXMEM_Data(EX_MUX_DataB, MEM_Data, 1'b1, reset, clk);
	Reg_1bit	EXMEM_MemToReg(EX_MemToReg, MEM_MemToReg, 1'b1, reset, clk);
	Reg_1bit	EXMEM_RegWrite(EX_RegWrite, MEM_RegWrite, 1'b1, reset, clk);
	Reg_1bit	EXMEM_MemWrite(EX_MemWrite, MEM_MemWrite, 1'b1, reset, clk);
	Reg_1bit	EXMEM_MemRead(EX_MemRead, MEM_MemRead, 1'b1, reset, clk);
	Reg_5bit	EXMEM_Rd(EX_Rd, MEM_Rd, 1'b1, reset, clk);	
	
	//MEMory
	Memory	DataMemory(MEM_ALUResult, MEM_Data, MEM_ReadData, MEM_MemWrite, MEM_MemRead, clk);
	
	// MEM/WB reg
	Reg_32bit	MEMWB_ReadData(MEM_ReadData,  WB_ReadData, 1'b1, reset, clk);
	Reg_32bit	MEMWB_Address(MEM_ALUResult, WB_Address, 1'b1, reset, clk);
	Reg_5bit	MEMWB_Rd(MEM_Rd,        WB_WriteAddr, 1'b1, reset, clk);	
	Reg_1bit 	MEMWB_MemToReg(MEM_MemToReg,  WB_MemToReg, 1'b1, reset, clk);
	Reg_1bit	MEMWB_RegWrite(MEM_RegWrite,  WB_RegWrite, 1'b1, reset, clk);
	
	//WB
	MUX2x1_32bit	WB_MUX(WB_MemToReg, WB_Address, WB_ReadData, WB_WriteData);
endmodule
